# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Airbyte CDK 'image' commands.

The `airbyte-cdk image build` command provides a simple way to work with Airbyte
connector images.
"""

import sys
from pathlib import Path

import rich_click as click

from airbyte_cdk.cli.airbyte_cdk._util import resolve_connector_name_and_directory
from airbyte_cdk.models.connector_metadata import MetadataFile
from airbyte_cdk.utils.docker import (
    ConnectorImageBuildError,
    build_connector_image,
    verify_docker_installation,
)


@click.group(
    name="image",
    help=__doc__.replace("\n", "\n\n"),  # Render docstring as help text (markdown)
)
def image_cli_group() -> None:
    """Commands for working with connector Docker images."""


@image_cli_group.command()
@click.option(
    "--connector-name",
    type=str,
    help="Name of the connector to test. Ignored if --connector-directory is provided.",
)
@click.option(
    "--connector-directory",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    help="Path to the connector directory.",
)
@click.option("--tag", default="dev", help="Tag to apply to the built image (default: dev)")
@click.option("--no-verify", is_flag=True, help="Skip verification of the built image")
def build(
    connector_name: str | None = None,
    connector_directory: Path | None = None,
    *,
    tag: str = "dev",
    no_verify: bool = False,
) -> None:
    """Build a connector Docker image.

    This command builds a Docker image for a connector, using either
    the connector's Dockerfile or a base image specified in the metadata.
    The image is built for both AMD64 and ARM64 architectures.
    """
    if not verify_docker_installation():
        click.echo(
            "Docker is not installed or not running. Please install Docker and try again.", err=True
        )
        sys.exit(1)

    connector_name, connector_directory = resolve_connector_name_and_directory(
        connector_name=connector_name,
        connector_directory=connector_directory,
    )

    metadata_file_path: Path = connector_directory / "metadata.yaml"
    try:
        metadata = MetadataFile.from_file(metadata_file_path)
    except (FileNotFoundError, ValueError) as e:
        click.echo(
            f"Error loading metadata file '{metadata_file_path}': {e!s}",
            err=True,
        )
        sys.exit(1)
    click.echo(f"Building Image for Connector: {metadata.data.dockerRepository}:{tag}")
    try:
        build_connector_image(
            connector_directory=connector_directory,
            connector_name=connector_name,
            metadata=metadata,
            tag=tag,
            no_verify=no_verify,
        )
    except ConnectorImageBuildError as e:
        click.echo(
            f"Error building connector image: {e!s}",
            err=True,
        )
        sys.exit(1)


__all__ = [
    "image_cli_group",
]



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""A collection of Dockerfile templates for building Airbyte connectors.

The templates are designed to be used with the Airbyte CDK and can be customized
for different connectors and architectures.

These templates are used to generate connector images.
"""

##############################
## GLOBAL DOCKERIGNORE FILE ##
##############################

DOCKERIGNORE_TEMPLATE: str = "\n".join(
    [
        "# This file is auto-generated. Do not edit.",
        # "*,"
        "build/",
        ".venv/",
        "secrets/",
        "!setup.py",
        "!pyproject.toml",
        "!poetry.lock",
        "!poetry.toml",
        "!components.py",
        "!requirements.txt",
        "!README.md",
        "!metadata.yaml",
        "!build_customization.py",
        "!source_*",
        "!destination_*",
    ]
)

###########################
# PYTHON CONNECTOR IMAGE ##
###########################

PYTHON_CONNECTOR_DOCKERFILE_TEMPLATE = """
# syntax=docker/dockerfile:1
# check=skip=all
ARG BASE_IMAGE

FROM ${BASE_IMAGE} AS builder
ARG BASE_IMAGE
ARG CONNECTOR_SNAKE_NAME
ARG CONNECTOR_KEBAB_NAME
ARG EXTRA_PREREQS_SCRIPT=""

WORKDIR /airbyte/integration_code

COPY . ./

# Conditionally copy and execute the extra build script if provided
RUN if [ -n "${EXTRA_PREREQS_SCRIPT}" ]; then \
        cp ${EXTRA_PREREQS_SCRIPT} ./extra_prereqs_script && \
        ./extra_prereqs_script; \
    fi

# TODO: Pre-install uv on the base image to speed up the build.
#       (uv is still faster even with the extra step.)
RUN pip install --no-cache-dir uv
RUN python -m uv pip install --no-cache-dir .

FROM ${BASE_IMAGE}
ARG CONNECTOR_SNAKE_NAME
ARG CONNECTOR_KEBAB_NAME
ARG BASE_IMAGE

WORKDIR /airbyte/integration_code

COPY --from=builder /usr/local /usr/local
COPY --chmod=755 <<EOT /entrypoint.sh
#!/usr/bin/env bash
set -e

${CONNECTOR_KEBAB_NAME} "\$\@"
EOT

ENV AIRBYTE_ENTRYPOINT="/entrypoint.sh"
ENTRYPOINT ["/entrypoint.sh"]
"""

##################################
# MANIFEST-ONLY CONNECTOR IMAGE ##
##################################

MANIFEST_ONLY_DOCKERFILE_TEMPLATE = """
ARG BASE_IMAGE
ARG CONNECTOR_SNAKE_NAME
ARG CONNECTOR_KEBAB_NAME

FROM ${BASE_IMAGE}

WORKDIR /airbyte/integration_code

COPY . ./

ENV AIRBYTE_ENTRYPOINT="python ./main.py"
ENTRYPOINT ["python", "./main.py"]
"""



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from dataclasses import asdict, dataclass, field
from typing import Any, ClassVar, Dict, List, Mapping

from airbyte_cdk.connector_builder.test_reader import TestReader
from airbyte_cdk.models import (
    AirbyteMessage,
    AirbyteRecordMessage,
    AirbyteStateMessage,
    ConfiguredAirbyteCatalog,
    Type,
)
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.declarative.declarative_source import DeclarativeSource
from airbyte_cdk.sources.declarative.manifest_declarative_source import ManifestDeclarativeSource
from airbyte_cdk.sources.declarative.parsers.model_to_component_factory import (
    ModelToComponentFactory,
)
from airbyte_cdk.utils.airbyte_secrets_utils import filter_secrets
from airbyte_cdk.utils.datetime_helpers import ab_datetime_now
from airbyte_cdk.utils.traced_exception import AirbyteTracedException

DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE = 5
DEFAULT_MAXIMUM_NUMBER_OF_SLICES = 5
DEFAULT_MAXIMUM_RECORDS = 100
DEFAULT_MAXIMUM_STREAMS = 100

MAX_PAGES_PER_SLICE_KEY = "max_pages_per_slice"
MAX_SLICES_KEY = "max_slices"
MAX_RECORDS_KEY = "max_records"
MAX_STREAMS_KEY = "max_streams"


@dataclass
class TestLimits:
    __test__: ClassVar[bool] = False  # Tell Pytest this is not a Pytest class, despite its name

    max_records: int = field(default=DEFAULT_MAXIMUM_RECORDS)
    max_pages_per_slice: int = field(default=DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE)
    max_slices: int = field(default=DEFAULT_MAXIMUM_NUMBER_OF_SLICES)
    max_streams: int = field(default=DEFAULT_MAXIMUM_STREAMS)


def get_limits(config: Mapping[str, Any]) -> TestLimits:
    command_config = config.get("__test_read_config", {})
    max_pages_per_slice = (
        command_config.get(MAX_PAGES_PER_SLICE_KEY) or DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE
    )
    max_slices = command_config.get(MAX_SLICES_KEY) or DEFAULT_MAXIMUM_NUMBER_OF_SLICES
    max_records = command_config.get(MAX_RECORDS_KEY) or DEFAULT_MAXIMUM_RECORDS
    max_streams = command_config.get(MAX_STREAMS_KEY) or DEFAULT_MAXIMUM_STREAMS
    return TestLimits(max_records, max_pages_per_slice, max_slices, max_streams)


def should_normalize_manifest(config: Mapping[str, Any]) -> bool:
    """
    Check if the manifest should be normalized.
    :param config: The configuration to check
    :return: True if the manifest should be normalized, False otherwise.
    """
    return config.get("__should_normalize", False)


def create_source(config: Mapping[str, Any], limits: TestLimits) -> ManifestDeclarativeSource:
    manifest = config["__injected_declarative_manifest"]
    return ManifestDeclarativeSource(
        config=config,
        emit_connector_builder_messages=True,
        source_config=manifest,
        normalize_manifest=should_normalize_manifest(config),
        component_factory=ModelToComponentFactory(
            emit_connector_builder_messages=True,
            limit_pages_fetched_per_slice=limits.max_pages_per_slice,
            limit_slices_fetched=limits.max_slices,
            disable_retries=True,
            disable_cache=True,
        ),
    )


def read_stream(
    source: DeclarativeSource,
    config: Mapping[str, Any],
    configured_catalog: ConfiguredAirbyteCatalog,
    state: List[AirbyteStateMessage],
    limits: TestLimits,
) -> AirbyteMessage:
    try:
        test_read_handler = TestReader(
            limits.max_pages_per_slice, limits.max_slices, limits.max_records
        )
        # The connector builder only supports a single stream
        stream_name = configured_catalog.streams[0].stream.name

        stream_read = test_read_handler.run_test_read(
            source, config, configured_catalog, state, limits.max_records
        )

        return AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(
                data=asdict(stream_read), stream=stream_name, emitted_at=_emitted_at()
            ),
        )
    except Exception as exc:
        error = AirbyteTracedException.from_exception(
            exc,
            message=filter_secrets(
                f"Error reading stream with config={config} and catalog={configured_catalog}: {str(exc)}"
            ),
        )
        return error.as_airbyte_message()


def resolve_manifest(source: ManifestDeclarativeSource) -> AirbyteMessage:
    try:
        return AirbyteMessage(
            type=Type.RECORD,
            record=AirbyteRecordMessage(
                data={"manifest": source.resolved_manifest},
                emitted_at=_emitted_at(),
                stream="resolve_manifest",
            ),
        )
    except Exception as exc:
        error = AirbyteTracedException.from_exception(
            exc, message=f"Error resolving manifest: {str(exc)}"
        )
        return error.as_airbyte_message()


def full_resolve_manifest(source: ManifestDeclarativeSource, limits: TestLimits) -> AirbyteMessage:
    try:
        manifest = {**source.resolved_manifest}
        streams = manifest.get("streams", [])
        for stream in streams:
            stream["dynamic_stream_name"] = None

        mapped_streams: Dict[str, List[Dict[str, Any]]] = {}
        for stream in source.dynamic_streams:
            generated_streams = mapped_streams.setdefault(stream["dynamic_stream_name"], [])

            if len(generated_streams) < limits.max_streams:
                generated_streams += [stream]

        for generated_streams_list in mapped_streams.values():
            streams.extend(generated_streams_list)

        manifest["streams"] = streams
        return AirbyteMessage(
            type=Type.RECORD,
            record=AirbyteRecordMessage(
                data={"manifest": manifest},
                emitted_at=_emitted_at(),
                stream="full_resolve_manifest",
            ),
        )
    except AirbyteTracedException as exc:
        return exc.as_airbyte_message()
    except Exception as exc:
        error = AirbyteTracedException.from_exception(
            exc, message=f"Error full resolving manifest: {str(exc)}"
        )
        return error.as_airbyte_message()


def _emitted_at() -> int:
    return ab_datetime_now().to_epoch_millis()



"""Common utilities for Airbyte CDK CLI."""

from pathlib import Path

from airbyte_cdk.test.standard_tests.test_resources import find_connector_root_from_name


def resolve_connector_name_and_directory(
    connector_name: str | None = None,
    connector_directory: Path | None = None,
) -> tuple[str, Path]:
    """Resolve the connector name and directory.

    This function will resolve the connector name and directory based on the provided
    arguments. If no connector name or directory is provided, it will look within the
    current working directory. If the current working directory is not a connector
    directory (e.g. starting with 'source-') and no connector name or path is provided,
    the process will fail.
    """
    if not connector_directory:
        if connector_name:
            connector_directory = find_connector_root_from_name(connector_name)
        else:
            cwd = Path().resolve().absolute()
            if cwd.name.startswith("source-") or cwd.name.startswith("destination-"):
                connector_directory = cwd
            else:
                raise ValueError(
                    "Either connector_name or connector_directory must be provided if not "
                    "running from a connector directory."
                )

    if not connector_name:
        connector_name = connector_directory.name

    if connector_directory:
        connector_directory = connector_directory.resolve().absolute()
    elif connector_name:
        connector_directory = find_connector_root_from_name(connector_name)
    else:
        raise ValueError("Either connector_name or connector_directory must be provided.")

    return connector_name, connector_directory



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Base class for connector test suites."""

from __future__ import annotations

import abc
import importlib
import inspect
import os
import sys
from collections.abc import Callable
from pathlib import Path
from typing import cast

import yaml
from boltons.typeutils import classproperty

from airbyte_cdk.models import (
    AirbyteMessage,
    Type,
)
from airbyte_cdk.test import entrypoint_wrapper
from airbyte_cdk.test.standard_tests._job_runner import IConnector, run_test_job
from airbyte_cdk.test.standard_tests.models import (
    ConnectorTestScenario,
)
from airbyte_cdk.test.standard_tests.test_resources import (
    ACCEPTANCE_TEST_CONFIG,
    find_connector_root,
)


class ConnectorTestSuiteBase(abc.ABC):
    """Base class for connector test suites."""

    connector: type[IConnector] | Callable[[], IConnector] | None  # type: ignore [reportRedeclaration]
    """The connector class or a factory function that returns an scenario of IConnector."""

    @classproperty  # type: ignore [no-redef]
    def connector(cls) -> type[IConnector] | Callable[[], IConnector] | None:
        """Get the connector class for the test suite.

        This assumes a python connector and should be overridden by subclasses to provide the
        specific connector class to be tested.
        """
        connector_root = cls.get_connector_root_dir()
        connector_name = connector_root.absolute().name

        expected_module_name = connector_name.replace("-", "_").lower()
        expected_class_name = connector_name.replace("-", "_").title().replace("_", "")

        # dynamically import and get the connector class: <expected_module_name>.<expected_class_name>

        cwd_snapshot = Path().absolute()
        os.chdir(connector_root)

        # Dynamically import the module
        try:
            module = importlib.import_module(expected_module_name)
        except ModuleNotFoundError as e:
            raise ImportError(f"Could not import module '{expected_module_name}'.") from e
        finally:
            # Change back to the original working directory
            os.chdir(cwd_snapshot)

        # Dynamically get the class from the module
        try:
            return cast(type[IConnector], getattr(module, expected_class_name))
        except AttributeError as e:
            # We did not find it based on our expectations, so let's check if we can find it
            # with a case-insensitive match.
            matching_class_name = next(
                (name for name in dir(module) if name.lower() == expected_class_name.lower()),
                None,
            )
            if not matching_class_name:
                raise ImportError(
                    f"Module '{expected_module_name}' does not have a class named '{expected_class_name}'."
                ) from e
            return cast(type[IConnector], getattr(module, matching_class_name))

    @classmethod
    def get_test_class_dir(cls) -> Path:
        """Get the file path that contains the class."""
        module = sys.modules[cls.__module__]
        # Get the directory containing the test file
        return Path(inspect.getfile(module)).parent

    @classmethod
    def create_connector(
        cls,
        scenario: ConnectorTestScenario,
    ) -> IConnector:
        """Instantiate the connector class."""
        connector = cls.connector  # type: ignore
        if connector:
            if callable(connector) or isinstance(connector, type):
                # If the connector is a class or factory function, instantiate it:
                return cast(IConnector, connector())  # type: ignore [redundant-cast]

        # Otherwise, we can't instantiate the connector. Fail with a clear error message.
        raise NotImplementedError(
            "No connector class or connector factory function provided. "
            "Please provide a class or factory function in `cls.connector`, or "
            "override `cls.create_connector()` to define a custom initialization process."
        )

    # Test Definitions

    def test_check(
        self,
        scenario: ConnectorTestScenario,
    ) -> None:
        """Run `connection` acceptance tests."""
        result: entrypoint_wrapper.EntrypointOutput = run_test_job(
            self.create_connector(scenario),
            "check",
            test_scenario=scenario,
        )
        conn_status_messages: list[AirbyteMessage] = [
            msg for msg in result._messages if msg.type == Type.CONNECTION_STATUS
        ]  # noqa: SLF001  # Non-public API
        assert len(conn_status_messages) == 1, (
            f"Expected exactly one CONNECTION_STATUS message. Got: {result._messages}"
        )

    @classmethod
    def get_connector_root_dir(cls) -> Path:
        """Get the root directory of the connector."""
        return find_connector_root([cls.get_test_class_dir(), Path.cwd()])

    @classproperty
    def acceptance_test_config_path(cls) -> Path:
        """Get the path to the acceptance test config file."""
        result = cls.get_connector_root_dir() / ACCEPTANCE_TEST_CONFIG
        if result.exists():
            return result

        raise FileNotFoundError(f"Acceptance test config file not found at: {str(result)}")

    @classmethod
    def get_scenarios(
        cls,
    ) -> list[ConnectorTestScenario]:
        """Get acceptance tests for a given category.

        This has to be a separate function because pytest does not allow
        parametrization of fixtures with arguments from the test class itself.
        """
        category = "connection"
        all_tests_config = yaml.safe_load(cls.acceptance_test_config_path.read_text())
        if "acceptance_tests" not in all_tests_config:
            raise ValueError(
                f"Acceptance tests config not found in {cls.acceptance_test_config_path}."
                f" Found only: {str(all_tests_config)}."
            )
        if category not in all_tests_config["acceptance_tests"]:
            return []
        if "tests" not in all_tests_config["acceptance_tests"][category]:
            raise ValueError(f"No tests found for category {category}")

        tests_scenarios = [
            ConnectorTestScenario.model_validate(test)
            for test in all_tests_config["acceptance_tests"][category]["tests"]
            if "iam_role" not in test["config_path"]
        ]
        connector_root = cls.get_connector_root_dir().absolute()
        for test in tests_scenarios:
            if test.config_path:
                test.config_path = connector_root / test.config_path
            if test.configured_catalog_path:
                test.configured_catalog_path = connector_root / test.configured_catalog_path

        return tests_scenarios



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from dataclasses import InitVar, dataclass
from enum import Enum
from typing import Any, Iterable, List, Mapping, Optional

from airbyte_cdk.sources.declarative.requesters.query_properties.strategies import GroupByKey
from airbyte_cdk.sources.declarative.requesters.query_properties.strategies.merge_strategy import (
    RecordMergeStrategy,
)
from airbyte_cdk.sources.types import Config, Record


class PropertyLimitType(Enum):
    """
    The heuristic that determines when the maximum size of the current chunk of properties and when a new
    one should be started.
    """

    characters = "characters"
    property_count = "property_count"


@dataclass
class PropertyChunking:
    """
    Defines the behavior for how the complete list of properties to query for are broken down into smaller groups
    that will be used for multiple requests to the target API.
    """

    property_limit_type: PropertyLimitType
    property_limit: Optional[int]
    record_merge_strategy: Optional[RecordMergeStrategy]
    parameters: InitVar[Mapping[str, Any]]
    config: Config

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._record_merge_strategy = self.record_merge_strategy or GroupByKey(
            key="id", config=self.config, parameters=parameters
        )

    def get_request_property_chunks(
        self, property_fields: Iterable[str], always_include_properties: Optional[List[str]]
    ) -> Iterable[List[str]]:
        if not self.property_limit:
            single_property_chunk = list(property_fields)
            if always_include_properties:
                single_property_chunk.extend(always_include_properties)
            yield single_property_chunk
            return
        current_chunk = list(always_include_properties) if always_include_properties else []
        chunk_size = 0
        for property_field in property_fields:
            # If property_limit_type is not defined, we default to property_count which is just an incrementing count
            # todo: Add ability to specify parameter delimiter representation and take into account in property_field_size
            property_field_size = (
                len(property_field)
                + 3  # The +3 represents the extra characters for encoding the delimiter in between properties
                if self.property_limit_type == PropertyLimitType.characters
                else 1
            )
            if chunk_size + property_field_size > self.property_limit:
                yield current_chunk
                current_chunk = list(always_include_properties) if always_include_properties else []
                chunk_size = 0
            current_chunk.append(property_field)
            chunk_size += property_field_size
        yield current_chunk

    def get_merge_key(self, record: Record) -> Optional[str]:
        return self._record_merge_strategy.get_group_key(record=record)



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.


from dataclasses import InitVar, dataclass, field
from typing import Any, Iterable, Mapping, Optional

from airbyte_cdk.sources.declarative.async_job.job import AsyncJob
from airbyte_cdk.sources.declarative.extractors.record_selector import RecordSelector
from airbyte_cdk.sources.declarative.partition_routers.async_job_partition_router import (
    AsyncJobPartitionRouter,
)
from airbyte_cdk.sources.declarative.retrievers.retriever import Retriever
from airbyte_cdk.sources.streams.core import StreamData
from airbyte_cdk.sources.types import Config, StreamSlice, StreamState
from airbyte_cdk.sources.utils.slice_logger import AlwaysLogSliceLogger


@dataclass
class AsyncRetriever(Retriever):
    config: Config
    parameters: InitVar[Mapping[str, Any]]
    record_selector: RecordSelector
    stream_slicer: AsyncJobPartitionRouter
    slice_logger: AlwaysLogSliceLogger = field(
        init=False,
        default_factory=lambda: AlwaysLogSliceLogger(),
    )

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._parameters = parameters

    @property
    def exit_on_rate_limit(self) -> bool:
        """
        Whether to exit on rate limit. This is a property of the job repository
        and not the stream slicer. The stream slicer is responsible for creating
        the jobs, but the job repository is responsible for managing the rate
        limits and other job-related properties.

        Note:
         - If the `creation_requester` cannot place / create the job - it might be the case of the RateLimits
         - If the `creation_requester` can place / create the job - it means all other requesters should successfully manage
           to complete the results.
        """
        job_orchestrator = self.stream_slicer._job_orchestrator
        if job_orchestrator is None:
            # Default value when orchestrator is not available
            return False
        return job_orchestrator._job_repository.creation_requester.exit_on_rate_limit  # type: ignore

    @exit_on_rate_limit.setter
    def exit_on_rate_limit(self, value: bool) -> None:
        """
        Sets the `exit_on_rate_limit` property of the job repository > creation_requester,
        meaning that the Job cannot be placed / created if the rate limit is reached.
        Thus no further work on managing jobs is expected to be done.
        """
        job_orchestrator = self.stream_slicer._job_orchestrator
        if job_orchestrator is not None:
            job_orchestrator._job_repository.creation_requester.exit_on_rate_limit = value  # type: ignore[attr-defined, assignment]

    @property
    def state(self) -> StreamState:
        """
        As a first iteration for sendgrid, there is no state to be managed
        """
        return {}

    @state.setter
    def state(self, value: StreamState) -> None:
        """
        As a first iteration for sendgrid, there is no state to be managed
        """
        pass

    def _get_stream_state(self) -> StreamState:
        """
        Gets the current state of the stream.

        Returns:
            StreamState: Mapping[str, Any]
        """

        return self.state

    def _validate_and_get_stream_slice_jobs(
        self, stream_slice: Optional[StreamSlice] = None
    ) -> Iterable[AsyncJob]:
        """
        Validates the stream_slice argument and returns the partition from it.

        Args:
            stream_slice (Optional[StreamSlice]): The stream slice to validate and extract the partition from.

        Returns:
            AsyncPartition: The partition extracted from the stream_slice.

        Raises:
            AirbyteTracedException: If the stream_slice is not an instance of StreamSlice or if the partition is not present in the stream_slice.

        """
        return stream_slice.extra_fields.get("jobs", []) if stream_slice else []

    def stream_slices(self) -> Iterable[Optional[StreamSlice]]:
        yield from self.stream_slicer.stream_slices()

    def read_records(
        self,
        records_schema: Mapping[str, Any],
        stream_slice: Optional[StreamSlice] = None,
    ) -> Iterable[StreamData]:
        # emit the slice_descriptor log message, for connector builder TestRead
        yield self.slice_logger.create_slice_log_message(stream_slice.cursor_slice)  # type: ignore

        stream_state: StreamState = self._get_stream_state()
        jobs: Iterable[AsyncJob] = self._validate_and_get_stream_slice_jobs(stream_slice)
        records: Iterable[Mapping[str, Any]] = self.stream_slicer.fetch_records(jobs)

        yield from self.record_selector.filter_and_transform(
            all_data=records,
            stream_state=stream_state,
            records_schema=records_schema,
            stream_slice=stream_slice,
        )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from airbyte_cdk.sources.declarative.requesters.query_properties.properties_from_endpoint import (
    PropertiesFromEndpoint,
)
from airbyte_cdk.sources.declarative.requesters.query_properties.property_chunking import (
    PropertyChunking,
)
from airbyte_cdk.sources.declarative.requesters.query_properties.query_properties import (
    QueryProperties,
)

__all__ = ["PropertiesFromEndpoint", "PropertyChunking", "QueryProperties"]
